import * as SAX from "sax";
const SAXParser = SAX.default;
function possibleBytesToString(str, textDecoder = new TextDecoder()) {
    return typeof str === "string" ? str : textDecoder.decode(str);
}
/**
 * @param options
 * @returns A through stream that can be async iterated. Will emit data from any of the events that are listened to in the form `{event: "event-name", payload: "event-data" }` where "event-name" and "event-data" are the event's name and data.
 */
export function streamingSAX({ saxOptions = {}, events, continueOnError = false, }) {
    let parser = null;
    const textDecoder = new TextDecoder();
    const transform = new TransformStream({
        start(controller) {
            const { strict, ...opts } = saxOptions;
            parser = SAXParser.parser(strict !== null && strict !== void 0 ? strict : false, opts);
            const ERROR_EVENT = "error";
            if (continueOnError && events.includes(ERROR_EVENT)) {
                parser.onerror = (payload) => {
                    const queueableEvent = {
                        event: ERROR_EVENT,
                        payload,
                    };
                    controller.enqueue(queueableEvent);
                    parser === null || parser === void 0 ? void 0 : parser.resume();
                };
                events = events.filter((x) => x === ERROR_EVENT);
            }
            else if (continueOnError) {
                parser.onerror = () => {
                    parser === null || parser === void 0 ? void 0 : parser.resume();
                };
            }
            for (const event of events) {
                parser[`on${event}`] = (...args) => {
                    const payload = args[0];
                    controller.enqueue((payload ? {
                        event,
                        payload,
                    } : { event }));
                };
            }
            // parser.onend = () => {
            //   controller.terminate();
            // };
        },
        transform(chunk) {
            parser === null || parser === void 0 ? void 0 : parser.write(possibleBytesToString(chunk, textDecoder));
        },
    });
    return transform;
}
/** */
function xmlStateToJSON(data) {
    var _a, _b, _c, _d, _e, _f, _g, _h, _j;
    return {
        [data.name]: ((_a = data.children) === null || _a === void 0 ? void 0 : _a.length) === 0 && !data.attributes
            ? null
            : ((_b = data.children) === null || _b === void 0 ? void 0 : _b.length) === 1 && ((_d = (_c = data.children) === null || _c === void 0 ? void 0 : _c[0]) === null || _d === void 0 ? void 0 : _d.type) === "text"
                ? (_g = (_f = (_e = data.children) === null || _e === void 0 ? void 0 : _e[0]) === null || _f === void 0 ? void 0 : _f.value) !== null && _g !== void 0 ? _g : null
                : {
                    ...Object.fromEntries(Object.entries((_h = data.attributes) !== null && _h !== void 0 ? _h : {}).map(([key, value]) => [`@${key}`, value])),
                    ...("children" in data
                        ? (_j = data.children) === null || _j === void 0 ? void 0 : _j.reduce((acc, el) => {
                            var _a;
                            return el.type === "text"
                                ? { ...acc, "#text": (_a = el.value) !== null && _a !== void 0 ? _a : "" }
                                : el.type === "element"
                                    ? el.name in acc
                                        ? {
                                            ...acc,
                                            [el.name]: [
                                                ...(Array.isArray(acc[el.name])
                                                    ? acc[el.name]
                                                    : [
                                                        acc[el.name],
                                                    ]),
                                                xmlStateToJSON(el)[el.name],
                                            ],
                                        }
                                        : { ...acc, ...xmlStateToJSON(el) }
                                    : acc;
                        }, {})
                        : {}),
                },
    };
}
function* pathsToString(el) {
    let path = [];
    let ancestor = el;
    while (ancestor) {
        path = [ancestor.name, ...path];
        yield [path, ancestor];
        ancestor = ancestor.parent;
    }
}
function* first(iter) {
    for (const item of iter) {
        yield item;
        // This is intentional to get only one item.
        break;
    }
}
function checkNamePath(name, el, getAncestors = false) {
    for (const [path, ancestor] of getAncestors
        ? pathsToString(el)
        : first(pathsToString(el))) {
        if (name.has(path.join("."))) {
            return true;
        }
    }
    return false;
}
/**
 * @param stream A ReadableStream of xml in string or Uint8Array format.
 * @param opts The following are the properties of opts.
 *
 * **emit** - decides which emit strategy to take. a string of value "leafs", "updating-root", "leafs-named", or "updating-named"..
 *
 * **name** - names to filter our XML by. only matters if emit is either "named-leaf" or "updated-named". It's value is either a string or string[].
 *
 * @tutorial
 * ## Emit Modes:
 *
 * The various emit modes defined. Each emit only happens when a closing xml tag is hit but
 * depending on the emit mode it may be the closing tag of a subelement.
 *
 * ### "updating-root"
 *
 * Emits the root xml element as json over and over slowly adding new items and subitems to it as they are parsed.
 *
 * @example
 *
 * import { streamXMLToJSON } from "../../index.ts";
 * import { getExampleData } from "../utils.ts";
 *
 * async function getTitle(stream: ReadableStream<Uint8Array | string>) {
 *     for await (
 *         const item of streamXMLToJSON(stream, {
 *             emit: "updating-root",
 *         })
 *     ) {
 *         if (item?.rss?.channel?.title) {
 *             return item?.rss?.channel?.title;
 *         }
 *     }
 *     return "";
 * }
 * const stream = await getExampleData(0);
 *
 * console.log(await getTitle(stream));
 *
 * @tutorial
 *
 * ### "updating-named"
 *
 * Alot like "updating-root" except you give it a specific name or path of the only items you want emitted.
 * Note that the path syntax does not care where the first element of the path exists in the xml
 * it will also treat all arrays as if their items were just objects and emit only those that match.
 *
 * @example
 * import { streamXMLToJSON } from "../../index.ts";
 * import { getExampleData } from "../utils.ts";
 * // Get just enough data about the channel and then output.
 * async function getChannel(
 *     stream: ReadableStream<string | Uint8Array>,
 * ) {
 *     // In rss entries are generally small so we can
 *     for await (
 *         const { channel } of streamXMLToJSON(stream, {
 *             emit: "updating-named",
 *             name: "channel",
 *         })
 *     ) {
 *         console.log(channel, channel.image);
 *         if (
 *             channel.title &&
 *             channel.image
 *         ) {
 *             return channel;
 *         }
 *     }
 * }
 * const stream = await getExampleData(0);
 * console.log(await getChannel(stream));
 *
 * @tutorial
 *
 * ### "leafs-named"
 *
 * Alot like "leafs", but uses the name option to emit only leaf nodes that either match an exact name or a path that will lead to that name.
 * Note that the path syntax does not care where the first element of the path exists in the xml
 * it will also treat all arrays as if their items were just objects and emit only those that match.
 *
 * @example
 *
 * import { streamXMLToJSON } from "../../index.ts";
 * import { getExampleData } from "../utils.ts";
 *
 * async function* getChannelImage(
 *     stream: ReadableStream<string | Uint8Array>,
 * ) {
 *     for await (
 *         const { image, ["itunes:image"]: itunesImage } of streamXMLToJSON<
 *             {
 *                 image?: {
 *                     link?: string;
 *                     url?: string;
 *                     title?: string;
 *                     description?: string;
 *                 };
 *                 ["itunes:image"]?: { "@href"?: string };
 *             }
 *         >(
 *             stream,
 *             {
 *                 emit: "leafs-named",
 *                 // Get the channel's image or itunes:image
 *                 name: ["channel.image", "channel.itunes:image"],
 *             },
 *         )
 *     ) {
 *         yield { image, itunesImage } as const;
 *     }
 * }
 * const stream = await getExampleData(0);
 * for await (const item of getChannelImage(stream)) {
 *     console.log(item.image ?? item.itunesImage);
 * }
 *
 * @tutorial
 * ### "leafs"
 *
 * The most verbose of all options and probably not what you want, but get's any and all combinations of parts and pieces of json possible using breadth first search.
 * Basically walks the whole XML tree and emits JSON as it goes.
 *
 * @example
 * import { streamXMLToJSON } from "../../index.ts";
 * import { getExampleData } from "../utils.ts";
 *
 * async function getRSSTitle(stream: ReadableStream<Uint8Array | string>) {
 *    for await (
 *        const item of streamXMLToJSON(stream, {
 *            emit: "updating-root",
 *        })
 *    ) {
 *        if (item?.rss?.channel?.title) {
 *            return item?.rss?.channel?.title;
 *        }
 *    }
 *    return "";
 * }
 * const stream = await getExampleData(0);
 * console.log(await getTitle(stream));
 */
export async function* streamXMLToJSON(stream, { ...opts } = {}) {
    var _a, _b, _c, _d;
    const { emit = "updating-root" } = opts;
    const validEmits = [
        "leafs",
        "updating-root",
        "leafs-named",
        "updating-named",
    ];
    console.assert(validEmits.includes(emit), `You provided an incorrect emit method please choose either ${validEmits.slice(0, -1).join(",")}, or ${validEmits.slice(-1).join("")}.`);
    const stack = [];
    let getAncestors = false;
    const names = typeof opts.name === "string" || Array.isArray(opts.name)
        ? [opts.name].flat(1)
        : [];
    if (names.some((x) => x.includes("."))) {
        getAncestors = true;
    }
    // Treat this as a read only set...
    const name_set = new Set(names);
    for await (const item of stream.pipeThrough(streamingSAX({
        saxOptions: {
            lowercase: true,
            trim: true,
        },
        events: ["opentag", "text", "cdata", "error", "closetag"],
    }))) {
        if (item.event === "error") {
            throw item.payload;
        }
        switch (item.event) {
            case "opentag": {
                const state = {
                    type: "element",
                    name: item.payload.name,
                    parent: (_a = stack[stack.length - 1]) !== null && _a !== void 0 ? _a : null,
                    ...(Object.keys(item.payload.attributes).length > 0
                        ? { attributes: item.payload.attributes }
                        : {}),
                };
                if (state.parent) {
                    const parentState = state.parent;
                    if (parentState && !Array.isArray(parentState.children)) {
                        parentState.children = [];
                    }
                    (_b = parentState === null || parentState === void 0 ? void 0 : parentState.children) === null || _b === void 0 ? void 0 : _b.push(state);
                }
                stack.push(state);
                break;
            }
            case "text": {
                const parentState = stack[stack.length - 1];
                if (parentState && !Array.isArray(parentState.children)) {
                    parentState.children = [];
                }
                (_c = parentState === null || parentState === void 0 ? void 0 : parentState.children) === null || _c === void 0 ? void 0 : _c.push({
                    type: "text",
                    value: item.payload,
                });
                break;
            }
            case "cdata": {
                const parentState = stack[stack.length - 1];
                if (parentState && !Array.isArray(parentState.children)) {
                    parentState.children = [];
                }
                (_d = parentState === null || parentState === void 0 ? void 0 : parentState.children) === null || _d === void 0 ? void 0 : _d.push({
                    type: "text",
                    value: item.payload,
                });
                break;
            }
            case "closetag": {
                const state = stack.pop();
                if (state) {
                    if (emit === "leafs" ||
                        opts.emit === "leafs-named" &&
                            checkNamePath(name_set, state, getAncestors)) {
                        yield xmlStateToJSON(state);
                    }
                    else if (emit === "updating-root") {
                        if (stack[0]) {
                            yield xmlStateToJSON(stack[0]);
                        }
                    }
                    if (opts.emit === "updating-named") {
                        // lookup all parents with names and emit them.
                        let ancestor = state;
                        while (ancestor) {
                            if (checkNamePath(name_set, ancestor, getAncestors)) {
                                yield xmlStateToJSON(ancestor);
                            }
                            ancestor = ancestor.parent;
                        }
                    }
                }
                break;
            }
        }
    }
}
